/*
 * citygml4j - The Open Source Java API for CityGML
 * https://github.com/citygml4j
 *
 * Copyright 2013-2019 Claus Nagel <claus.nagel@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.citygml4j.model.gml.geometry.aggregates;

import org.citygml4j.builder.copy.CopyBuilder;
import org.citygml4j.geometry.BoundingBox;
import org.citygml4j.model.common.base.ModelObjects;
import org.citygml4j.model.common.child.ChildList;
import org.citygml4j.model.common.visitor.GMLFunctor;
import org.citygml4j.model.common.visitor.GMLVisitor;
import org.citygml4j.model.common.visitor.GeometryFunctor;
import org.citygml4j.model.common.visitor.GeometryVisitor;
import org.citygml4j.model.gml.GMLClass;
import org.citygml4j.model.gml.geometry.primitives.AbstractSurface;
import org.citygml4j.model.gml.geometry.primitives.SurfaceArrayProperty;
import org.citygml4j.model.gml.geometry.primitives.SurfaceProperty;

import java.util.Arrays;
import java.util.List;

public class MultiSurface extends AbstractGeometricAggregate {
	private List<SurfaceProperty> surfaceMember;
	private SurfaceArrayProperty surfaceMembers;
	
	public MultiSurface() {
		
	}
	
	public MultiSurface(List<? extends AbstractSurface> abstractSurfaces) {
		for (AbstractSurface abstractSurface : abstractSurfaces)
			addSurfaceMember(new SurfaceProperty(abstractSurface));
	}
	
	public MultiSurface(AbstractSurface... abstractSurfaces) {
		this(Arrays.asList(abstractSurfaces));
	}
	
	public void addSurfaceMember(SurfaceProperty surfaceMember) {
		getSurfaceMember().add(surfaceMember);
	}

	public List<SurfaceProperty> getSurfaceMember() {
		if (surfaceMember == null)
			surfaceMember = new ChildList<>(this);
		
		return surfaceMember;
	}

	public SurfaceArrayProperty getSurfaceMembers() {
		return surfaceMembers;
	}

	public boolean isSetSurfaceMember() {
		return surfaceMember != null && !surfaceMember.isEmpty();
	}

	public boolean isSetSurfaceMembers() {
		return surfaceMembers != null;
	}

	public void setSurfaceMember(List<SurfaceProperty> surfaceMember) {
		this.surfaceMember = new ChildList<>(this, surfaceMember);
	}

	public void setSurfaceMembers(SurfaceArrayProperty surfaceMembers) {
		this.surfaceMembers = ModelObjects.setParent(surfaceMembers, this);
	}

	public void unsetSurfaceMember() {
		surfaceMember = ModelObjects.setNull(surfaceMember);
	}

	public boolean unsetSurfaceMember(SurfaceProperty surfaceMember) {
		return isSetSurfaceMember() && this.surfaceMember.remove(surfaceMember);
	}

	public void unsetSurfaceMembers() {
		surfaceMembers = ModelObjects.setNull(surfaceMembers);
	}

	public BoundingBox calcBoundingBox() {
		BoundingBox bbox = new BoundingBox();
		
		if (isSetSurfaceMember()) {
			for (SurfaceProperty surfaceProperty : getSurfaceMember())
				if (surfaceProperty.isSetSurface())
					bbox.update(surfaceProperty.getSurface().calcBoundingBox());
		}

		if (isSetSurfaceMembers()) {
			SurfaceArrayProperty surfaceArrayProperty = getSurfaceMembers();

			if (surfaceArrayProperty.isSetSurface())
				for (AbstractSurface abstractSurface : surfaceArrayProperty.getSurface())
					bbox.update(abstractSurface.calcBoundingBox());
		}
		
		return bbox;
	}

	public GMLClass getGMLClass() {
		return GMLClass.MULTI_SURFACE;
	}

	public Object copy(CopyBuilder copyBuilder) {
		return copyTo(new MultiSurface(), copyBuilder);
	}

	@Override
	public Object copyTo(Object target, CopyBuilder copyBuilder) {
		MultiSurface copy = (target == null) ? new MultiSurface() : (MultiSurface)target;
		super.copyTo(copy, copyBuilder);
		
		if (isSetSurfaceMember()) {
			for (SurfaceProperty part : surfaceMember) {
				SurfaceProperty copyPart = (SurfaceProperty)copyBuilder.copy(part);
				copy.addSurfaceMember(copyPart);
				
				if (part != null && copyPart == part)
					part.setParent(this);
			}
		}
		
		if (isSetSurfaceMembers()) {
			copy.setSurfaceMembers((SurfaceArrayProperty)copyBuilder.copy(surfaceMembers));
			if (copy.getSurfaceMembers() == surfaceMembers)
				surfaceMembers.setParent(this);
		}
		
		return copy;
	}
	
	public void accept(GeometryVisitor visitor) {
		visitor.visit(this);
	}

	public <T> T accept(GeometryFunctor<T> visitor) {
		return visitor.apply(this);
	}
	
	public void accept(GMLVisitor visitor) {
		visitor.visit(this);
	}
	
	public <T> T accept(GMLFunctor<T> visitor) {
		return visitor.apply(this);
	}

}
